#include "stdinc.hpp"
#include <math.h>


const int screenX= 1024, screenY= 768;
const int blockSize= 8;
const int srcWidth= ((screenX)/blockSize)+1, srcHeight= ((screenY-128)/blockSize)+1;
const int yOffset= 64/blockSize;




struct BlackDot
{
	float x, y;
	float vx,vy;
};

struct BrightFlash
{
	float x, y;
	float timeAlive;
	float GetStrength()
	{
		return (5- timeAlive)*200;
	}
};


#pragma pack( push, 1 )
struct TgaHeader
{
    byte  identsize;          // size of ID field that follows 18 byte header (0 usually)
    byte  colourmaptype;      // type of colour map 0=none, 1=has palette
    byte  imagetype;          // type of image 0=none,1=indexed,2=rgb,3=grey,+8=rle packed

    short colourmapstart;     // first colour map entry in palette
    short colourmaplength;    // number of colours in palette
    byte  colourmapbits;      // number of bits per palette entry 15,16,24,32

    short xstart;             // image x origin
    short ystart;             // image y origin
    short width;              // image width in pixels
    short height;             // image height in pixels
    byte  bits;               // image bits per pixel 8,16,24,32
    byte  descriptor;         // image descriptor bits (vh flip bits)
};
#pragma pack( pop )


struct Sprite
{
	int width, height;
	qbyte* data;

	void RenderTo( SDL_Surface& surf, int ox, int oy )
	{
		qbyte* pDest= (qbyte*)surf.pixels;
		for( int y= 0; y< height; y++ )
			for( int x= 0; x< width; x++ )
			{
				qbyte d= data[ x + y*width ];
				int alpha= d>>24;
				if( alpha )
				{
					int alphaI= 255-alpha;
					qbyte* dest= &pDest[ ox+x + (oy+y)*(surf.pitch/4) ];

					int
						c= ((d&0xff)*alpha + ((*dest)&0xff)*alphaI)>>8;
						
					*dest= c | (c<<8) | (c<<16);
				}
			}

	}

	void Load( const char* filename )
	{
		int fileSize= 0;
		byte* fileData= NULL;
		{
			FILE* file= fopen( filename, "rb" );

			if( file!= NULL )
			{
				fseek( file, 0, SEEK_END );
				fileSize= ftell( file );
				fseek( file, 0, SEEK_SET );

				fileData= new byte[ fileSize ];

				fread( fileData, 1, fileSize, file );

				fclose( file );
			}
			else
			{
				assert( false );
				return;
			}
		}

		if( fileData!= NULL )
		{
			qbyte * imageData= (qbyte*)&fileData[ sizeof( TgaHeader ) ];

			{
				TgaHeader* head= (TgaHeader*)fileData;
				if( ( head->identsize != 0 ) ||
					( head->bits!=32 ) ||
					( head->colourmaptype!= 0 ) ||
					( head->imagetype!= 2 ) || 
					( head->xstart!= 0 ) || 
					( head->ystart!= 0 ) )
				{
					// wrong format
					assert( false );
					return;
				}
				width= head->width;
				height= head->height;
			}

			data= new qbyte[ width*height ];

			memcpy( data, imageData, sizeof( qbyte )*width*height );

			delete fileData;
		}
	}
};

enum Mode
{
	startingLevelMode,
	playingLevelMode,
	endingLevelMode,
	loosingMode,
	showingInstructionsMode,
	titleMode
};

class Game
{
public:

	byte smallBuf[ srcWidth * srcHeight ];
	float floatBuf[ srcWidth * srcHeight ];
	float totalTime;
	struct SDL_Surface *mainSurface;
	bool quit;
	Timer timer;
	bool lmbDown;
	bool fullscreen;

	int level;

	Array< BlackDot > blackDots;
	Array< BrightFlash > brightFlashes;

	float mouseX, mouseY;

	float timeLeft, pointsGained; 

	Sprite timeSprite, scoreSprite,
		level1Sprite,
		level2Sprite,
		level3Sprite,
		level4Sprite,
		background1Sprite,
		background2Sprite,
		background3Sprite,
		background4Sprite,
		instructionsSprite,
		lostSprite,
		titleSprite,
		finishedSprite;

	Mode mode;
	float modeTime;

	float TimeAllowed()
	{
		switch( level )
		{
		case 0: return 30;
		case 1: return 30;
		case 2: return 30;
		case 3: return 3;
		}
		return 1;
	}
	float PointsNeeded()
	{
		switch( level )
		{
		case 0: return 5;
		case 1: return 8;
		case 2: return 10;
		case 3: return 10;
		}
		return 1;
	}

	Sprite& LevelSprite()
	{
		switch( level )
		{
		case 0:
			return level1Sprite;
		case 1:
			return level2Sprite;
		case 2:
			return level3Sprite;
		case 3:
			return level4Sprite;
		};

		assert( false );
		return level1Sprite;
	}

	Sprite& BackgroundSprite()
	{
		switch( level )
		{
		case 0:
			return background1Sprite;
		case 1:
			return background2Sprite;
		case 2:
			return background3Sprite;
		case 3:
			return background4Sprite;
		};

		assert( false );
		return level1Sprite;
	}


	float GravityStrength()
	{
		switch( level )
		{
		case 0: return 10;
		case 1: return 10;
		case 2: return 20;
		case 3: return 20;
		};

		assert( false );
		return 10;
	}


	float GravityStrengthStrong()
	{
		switch( level )
		{
		case 0: return 40;
		case 1: return 60;
		case 2: return 80;
		case 3: return 80;
		};

		assert( false );
		return 10;
	}

	float Motes()
	{
		switch( level )
		{
		case 0: return 1;
		case 1: return 2;
		case 2: return 2;
		case 3: return 2;
		};

		assert( false );
		return 1;
	}

	Game()
	{
		fullscreen= true;
		totalTime= 0;
		quit= false;
		lmbDown= false;
		level= 0;

		timeSprite.Load( "data\\time.tga" );
		scoreSprite.Load( "data\\score.tga" );
		level1Sprite.Load( "data\\level1.tga" );
		level2Sprite.Load( "data\\level2.tga" );
		level3Sprite.Load( "data\\level3.tga" );
		level4Sprite.Load( "data\\level4.tga" );
		titleSprite.Load( "data\\title.tga" );
		finishedSprite.Load( "data\\finished.tga" );

		background1Sprite.Load( "data\\back1.tga" );
		background2Sprite.Load( "data\\back2.tga" );
		background3Sprite.Load( "data\\back3.tga" );
		background4Sprite.Load( "data\\back4.tga" );

		instructionsSprite.Load( "data\\instructions.tga" );
		lostSprite.Load( "data\\lost.tga" );


		modeTime= 0;
		mode= titleMode;
	}

	void StartGame()
	{
		level= 0;
		NewLevel();
	}

	void NewLevel()
	{
		blackDots.RemoveAll();
		pointsGained= 0;
		timeLeft= TimeAllowed();

		blackDots.RemoveAll();
		for( int i= 0; i< Motes(); i++ )
		{
			BlackDot& bd= blackDots.Insert();
			
			bd.x= Random01()*srcWidth;
			bd.y= Random01()*srcHeight;
			bd.vx= 0;
			bd.vy= 0;
		}

		modeTime= 0;
		mode= startingLevelMode;
	}

	void ClearScreen()
	{
		memset( mainSurface->pixels, 0, mainSurface->pitch*mainSurface->h );
	}

	void DoTime( const float timeSlice )
	{
		totalTime+= timeSlice;

		switch( mode )
		{
		case startingLevelMode:
			{
				modeTime+= timeSlice;

				if( modeTime > 2 )
				{
					mode= playingLevelMode;
				}
			}
			break;

		case endingLevelMode:
			{
				modeTime+= timeSlice;

				if( modeTime > 2 )
				{
					level++;

					if( level > 3 )
						mode= titleMode;
					else
						NewLevel();
				}
			}
			break;

		case loosingMode:
			{
				modeTime+= timeSlice;

				if( modeTime > 2 )
				{
					mode= titleMode;
				}
			}
			break;

		case playingLevelMode:
			{
				timeLeft-= timeSlice;

				if( timeLeft < 0 )
				{
					mode= loosingMode;
					modeTime= 0;
				}

				if( pointsGained > PointsNeeded() )
				{
					mode= endingLevelMode;
					modeTime= 0;
				}

				for( int i= 0; i< blackDots.size; i++ )
				{
					BlackDot& bd= blackDots[ i ];

					float 
						dx= mouseX - bd.x,
						dy= mouseY - bd.y,
						d= sqrt( dx*dx + dy*dy ),
						ndx= dx/d,
						ndy= dy/d;
						

					bd.x+= bd.vx*timeSlice;
					bd.y+= bd.vy*timeSlice;

					bd.vx*= (1.0-timeSlice*0.1);
					bd.vy*= (1.0-timeSlice*0.1);

					float gs= lmbDown ? GravityStrengthStrong() : GravityStrength();
					bd.vx+= (gs*ndx) *timeSlice;
					bd.vy+=(gs*ndy) *timeSlice;


					{
						float
							mx= (srcWidth/2.0) - bd.x, 
							my= (srcHeight/2.0) - bd.y, 
							md= sqrt( mx*mx + my*my );

						if( md < 20 )
						{
							if( md < 15 )
							{
								if( md < 5 )
									pointsGained-= timeSlice*3;
								else
									pointsGained-= timeSlice;
							}
							else
								pointsGained+= timeSlice*2;
						}

						if( pointsGained < 0 )
							pointsGained= 0;
					}
				}



				{
					bool skip;
					do
					{
						skip= false;
						for( int i= 0; (i< brightFlashes.size)&&!skip; i++ )
						{
							BrightFlash &bf= brightFlashes[ i ];

							if( bf.timeAlive > 5 )
							{
								brightFlashes.Remove( i );
								skip= true;
							}
						}
					}
					while( skip );
				}
			}
			break;
		case titleMode:
		default:
			break;
		};


	}


	void ExpandConvert( byte* src, qbyte* dest, int destPitch )
	{
		const int tempSize= screenX;
		short int temp[ screenX ];
/*
		for( int yy= 0; yy < srcHeight-1; yy++ )
			for( int xx= 0; xx < srcWidth-1; xx++ )
				for( int y= 0; y < blockSize; y++ )
					for( int x= 0; x < blockSize; x++ )
					{
						qbyte* pDest= &dest[ (xx*blockSize + x) + (
							(yy+yOffset)*blockSize + y)*destPitch ];
						byte c= src[ xx+ yy*srcWidth ];

						*pDest= (qbyte)( c | (c<<8) | (c<<16) );
					}

*/

		qbyte *destPtr= &dest[ yOffset*8*destPitch ];
		
		for( int blockY= 0; blockY< (srcHeight)-1; blockY++ )
		{
			{
				// prepare temp buffer
				byte* srcPtr= &src[ blockY*srcWidth ];
				
				for( int tx= 0; tx< srcWidth; tx++ )
				{
					short int 
						r= srcPtr[ tx ]<<7, 
						nr= srcPtr[ tx + srcWidth ] <<7;
					temp[ tx*2 + 0 ]= r;
					temp[ tx*2 + 1 ]= (nr - r)/blockSize;
				}
			}
			
			for( int y= 0; y< blockSize; y++ )
			{
				short int * tempPtr= temp;

				for( int blockX= 0; blockX< srcWidth-1; blockX++ )
				{
					float 
						currColor= (*tempPtr)/128.0,
						deltaColor= 
							((tempPtr[ 2 ] - *tempPtr)/(float)blockSize)/
							128.0;

					for( int x= 0; x< blockSize; x++ )
					{
						int c= currColor;

						*destPtr= c | (c<<8) | (c<<16);
						currColor+= deltaColor;
						destPtr++;
					}

					tempPtr[0]+= tempPtr[1];

					tempPtr+= 2;
				}
				tempPtr[0]+= tempPtr[1];
			}
		}

	}

	void RenderFlarePointy( float fx, float fy, float strength )
	{
		for( int y= 0; y< srcHeight; y++ )
			for( int x= 0; x< srcWidth; x++ )
			{
				float dx= fabs( fx-x ), dy= fabs( fy-y );
				floatBuf[ x + y*srcWidth ]+= strength/(dx*dy);
			}
	}

	void RenderFlare( float fx, float fy, float strength )
	{
		for( int y= 0; y< srcHeight; y++ )
			for( int x= 0; x< srcWidth; x++ )
			{
				float dx= fabs( fx-x ), dy= fabs( fy-y );
				float dist= sqrt( dx*dx + dy*dy );
				if( dist > 0 )
					floatBuf[ x + y*srcWidth ]+= strength/dist;
				else
					floatBuf[ x + y*srcWidth ]+= strength;
			}
	}

	void RenderCircle( float fx, float fy, float strength )
	{
		for( int y= 0; y< srcHeight; y++ )
			for( int x= 0; x< srcWidth; x++ )
			{
				float dx= fabs( fx-x ), dy= fabs( fy-y );
				float dist= sqrt( dx*dx + dy*dy );
					
				dist = fabs( dist - 15 );

				if( dist > 0 )
					floatBuf[ x + y*srcWidth ]+= strength/dist;
				else
					floatBuf[ x + y*srcWidth ]+= strength;
			}
	}

	void RenderTimeBar()
	{
		qbyte* pBuf= (qbyte*)mainSurface->pixels;

		int sz= (timeLeft/TimeAllowed())*(screenX-16);

		for( int y= 0; y< 48; y++ )
			for( int x= 0; x< (screenX-16); x++ )
			{
				qbyte c= x<sz ? 0xffffff : 0x0;
				pBuf[ (x + 8) + (mainSurface->pitch/4)*(y+8) ]= c;
			}
	}

	void RenderPointsBar()
	{
		qbyte* pBuf= (qbyte*)mainSurface->pixels;

		int sz= (pointsGained/PointsNeeded())*(screenX-16);

		for( int y= 0; y< 48; y++ )
			for( int x= 0; x< (screenX-16); x++ )
			{
				qbyte c= x<sz ? 0xffffff : 0x0;
				pBuf[ (x + 8) + (mainSurface->pitch/4)*(y+screenY- 48 -8) ]= c;
			}
	}

	void Render()
	{
		qbyte* pBuf= (qbyte*)mainSurface->pixels;


		if( (mode==playingLevelMode ) ||
			(mode==startingLevelMode ) ||
			(mode==endingLevelMode ) ||
			(mode==loosingMode )
			)
			{

				{
					Sprite& bs= BackgroundSprite();

					for( int i= 0; i< srcWidth*srcHeight; i++ )
						floatBuf[ i ]= bs.data[ i ] & 0xff;
				}
			

				RenderFlare( 
					mouseX, 
					mouseY, lmbDown ? 1000 : 500 );

				RenderCircle( 
					srcWidth/2, 
					srcHeight/2, sin(totalTime*2)*100 + 250 );

		//		if( lmbDown )
		//			RenderFlarePointy( 
		//				mouseX/8.0, 
		//				mouseY/8.0, 1000 );

				for( int i= 0; i< blackDots.size; i++ )
				{
					BlackDot& bd= blackDots[ i ];
					RenderFlare( bd.x, bd.y, -200 );
				}

				for( int i= 0; i< brightFlashes.size; i++ )
				{
					BrightFlash& bf= brightFlashes[ i ];
					RenderFlare( bf.x, bf.y, 200 );
				}

				for( int i= 0; i< srcWidth*srcHeight; i++ )
				{
					float f= floatBuf[ i ];

					if( f < 0 )
						f= 0;
					else if( f > 255 )
						f= 255;

					smallBuf[ i ]= (byte)f;
				}

				ExpandConvert( 
					smallBuf, pBuf,
					mainSurface->pitch/4 );

				RenderTimeBar();
				RenderPointsBar();

				timeSprite.RenderTo( *mainSurface, 20, 30 );
				scoreSprite.RenderTo( *mainSurface, 20, 690 );
			}

		if( mode == startingLevelMode )
		{
			Sprite& levelSprite= LevelSprite();
			if( ((int)(totalTime*3))%2 )
				levelSprite.RenderTo( *mainSurface, 
					screenX/2 - levelSprite.width/2,
					screenY/2 - levelSprite.height/2 );
		}

		if( mode == loosingMode )
		{
			if( ((int)(totalTime*3))%2 )
				lostSprite.RenderTo( *mainSurface, 
					screenX/2 - lostSprite.width/2,
					screenY/2 - lostSprite.height/2 );
		}

		if( mode == endingLevelMode )
		{
			Sprite& levelSprite= LevelSprite();

			if( ((int)(totalTime*3))%2 )
			{
				finishedSprite.RenderTo( *mainSurface, 
					screenX/2 - finishedSprite.width/2,
					screenY/2 - levelSprite.height/2 -40 );

				levelSprite.RenderTo( *mainSurface, 
					screenX/2 - levelSprite.width/2,
					screenY/2 - levelSprite.height/2 +40 );
			}
		}

		if( (mode == titleMode) || (mode==showingInstructionsMode) )
		{
			ClearScreen();
			for( int i= 0; i< srcWidth*srcHeight; i++ )
				floatBuf[ i ]= background4Sprite.data[ i ] & 0xff;
			for( int i= 0; i< srcWidth*srcHeight; i++ )
			{
				float f= floatBuf[ i ];

				if( f < 0 )
					f= 0;
				else if( f > 255 )
					f= 255;

				smallBuf[ i ]= (byte)f;
			}
			ExpandConvert( 
				smallBuf, pBuf,
				mainSurface->pitch/4 );

			if( mode==titleMode )
				titleSprite.RenderTo( *mainSurface, 
					screenX/2 - titleSprite.width/2,
					screenY/2 - titleSprite.height/2 );
			else
				instructionsSprite.RenderTo( *mainSurface, 
					screenX/2 - instructionsSprite.width/2,
					screenY/2 - instructionsSprite.height/2 );
		}
	}

	void NewKey( const SDL_keysym& ks )
	{
	}

	void MouseMove()
	{
	}

	void SetResolution( const int nx, const int ny )
	{
		mainSurface= SDL_SetVideoMode( nx, ny, 32, 
			(fullscreen ? SDL_FULLSCREEN : 0) 
			| SDL_HWSURFACE | SDL_DOUBLEBUF );

		if( ( mainSurface == NULL ) && fullscreen )
		{
			mainSurface= SDL_SetVideoMode( nx, ny, 32, 
				SDL_HWSURFACE | SDL_DOUBLEBUF );

			if( mainSurface == NULL )
				quit= true;

			fullscreen= false;
		}

		if( (mainSurface->pitch % 4 ) != 0 )
			quit= true;

		if( mainSurface->format->BitsPerPixel != 32 ) 
			quit= true;

		if( mainSurface->w != nx ) 
			quit= true;

		if( mainSurface->h != ny )
			quit= true;


		//display.ownedData= false;
		//display.size.x= nx;
		//display.size.y= ny;
		//display.pitch= mainSurface->pitch/sizeof(Pixel);
		//display.data= (Pixel*)mainSurface->pixels;
	}

	void Run()
	{
		SDL_Init( SDL_INIT_VIDEO );

		// -1 to detect bugs in drawing
		SetResolution( screenX, screenY );

		float totalTime= 0.0;

		float lastFrame= 0;
		
		while( !quit )
		{
			SDL_Event event;


			if( SDL_PollEvent( &event ) == 1 )
			{
				switch ( event.type )
				{
				case SDL_QUIT:
					quit= true;
					break;

				case SDL_KEYDOWN:
					{
						SDL_KeyboardEvent* e= (SDL_KeyboardEvent*)&event;

						if( ( e->keysym.sym == SDLK_RETURN ) &&
							( e->keysym.mod & KMOD_ALT) )
						{
							fullscreen= !fullscreen;
							SetResolution( screenX, screenY );
						}

						if( e->keysym.sym == 27 )
							quit= true;
					}
					break;

				case SDL_MOUSEMOTION:
					{
						SDL_MouseMotionEvent* e= (SDL_MouseMotionEvent*)&event;

						if( ( mode != loosingMode ) )
						{
							mouseX= (e->x)/8.0;
							mouseY= (e->y)/8.0  - yOffset;
						}
						MouseMove();
					}
					break;

				case SDL_MOUSEBUTTONUP:
					{
						SDL_MouseButtonEvent* e= (SDL_MouseButtonEvent*)&event;
						if( e->button == SDL_BUTTON_LEFT )
							lmbDown= false;

					}
					break;
				case SDL_MOUSEBUTTONDOWN:
					{
						SDL_MouseButtonEvent* e= (SDL_MouseButtonEvent*)&event;
						if( e->button == SDL_BUTTON_LEFT )
							lmbDown= true;

						if( mode== showingInstructionsMode )
							StartGame();
						if( mode== titleMode )
							mode= showingInstructionsMode;

					}
					break;
				};
			}
			else
			{
				float frameTime= timer.Tick();
				totalTime+= frameTime;

				lastFrame+= frameTime;
//				odsm( frameTime );
				
				DoTime( frameTime );

				if( lastFrame > 0.05 )
				{
					SDL_LockSurface( mainSurface );
					Render();
					SDL_UnlockSurface( mainSurface );

					SDL_Flip( mainSurface );

					lastFrame= 0;
				}
			}
		}		


		SDL_Quit();
	}
};

int WINAPI WinMain(
	HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, 
	int nCmdShow )
{

	Game game;

	game.Run();
	
	return 0;
}